"""
Proxy base class for lazy loading of objects.
Can be used as base class where a persistent mutable object is required to implement the methods of a given interface that consumes an external data source.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Iterable, TypeVar, cast

T = TypeVar("T")


class LazyProxy(Generic[T], ABC):
    """A lazy proxy class that defers the loading of the proxied object until it is accessed."""

    def __init__(self) -> None:
        self.__proxied: T | None = None

    def __getattr__(self, attr: str) -> object:
        return getattr(self.__get_proxied__(), attr)

    def __repr__(self) -> str:
        return repr(self.__get_proxied__())

    def __str__(self) -> str:
        return str(self.__get_proxied__())

    def __dir__(self) -> Iterable[str]:
        return self.__get_proxied__().__dir__()

    def __get_proxied__(self) -> T:
        proxied = self.__proxied
        if proxied is not None:
            return proxied

        self.__proxied = proxied = self.__load__()
        return proxied

    def __set_proxied__(self, value: T) -> None:
        self.__proxied = value

    def __as_proxied__(self) -> T:
        return cast(T, self)

    @abstractmethod
    def __load__(self) -> T: ...
